import logging
from enum import Enum
from typing import List
from .integer import Int
from .data_type import DataType

logger = logging.getLogger(__name__)


class Enums(DataType):

    def __init__(self, size: int, enum_class: callable, reverse: bool = False, default=None, decode_with_int=False):
        super().__init__(size, reverse, default)
        self._enum_class = enum_class
        self._decode_with_int = decode_with_int
        self._enum_list = [e for e in self._enum_class]

    def encode(self):
        result = [] if self.get_value() is None else self.get_value().value
        if not len(result) == self.get_size():
            result = result + [0] * (self.get_size() - len(result))
        return list(reversed(result)) if self._reverse else result

    def decode(self, array: List):
        array = list(reversed(array)) if self._reverse else array
        if self._decode_with_int:
            int_val = Int(self.get_size())
            int_val.decode(array)
            array = [int_val.get_value()]
        self.set_value(self._enum_class(array))

    def _default_value(self):
        return self._enum_list[0]

    def _value_check(self, value: Enum):
        if not type(value) == Enum:
            self.set_value(self._default_value())
            logger.error(f"{value} is not of type Enum. Using default value {self._default_value()}.")
            return False

        if value not in self._enum_list:
            logger.error(f"{value} doesn't belongs to enum class {self._enum_class}."
                         f" Using default value {self.default_value()}.")
            return False

        return True



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from ..data_types.array import Array
from ..data_types.enums import Enums
from ..data_types.integer import Int
from ..data_types.binary import Binary
from .command_packet import CommandPacket
from ..enums.adpd_enums import ADPDDevice, ADPDSlot, Clock, ADPDLed, ADPDAppID


class ActiveSlotPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xC',
                'checksum': '0x0'
            },
            'payload': {
                'command': <ADPDCommand.SET_SLOT_ACTIVE_RES: ['0x6B']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'slot_num': <ADPDSlot.SLOT_A: ['0x01']>,
                'slot_enabled': False
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["slot_num"] = Enums(1, enum_class=ADPDSlot)
        self._config["payload"]["slot_enabled"] = Binary(1)


class AgcControlPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xF',
                'checksum': '0x0'
            },
            'payload': {
                'command': <ADPDCommand.AGC_ON_OFF_RES: ['0x77']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'size': 2,
                'data': [
                    [ <ADPDLed.LED_MWL: ['0x00']>, True ],
                    [ <ADPDLed.LED_GREEN: ['0x01']>, True ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = Int(1)
        self._config["payload"]["data"] = Array(-1, dimension=2, data_types=[Binary(1), Enums(1, enum_class=ADPDLed)],
                                                reverse_inner_array=True)


class AgcInfoPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x61',
                'checksum': '0x0'
            },
            'payload': {
                'command': <ADPDCommand.AGC_INFO_RES: ['0x7B']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'led_index': <ADPDLed.LED_GREEN: ['0x01']>,
                'ch1': [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
                'ch2': [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 ],
                'dc0_led_current': 0,
                'tia_ch1': 0,
                'tia_ch2': 0
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["led_index"] = Enums(1, enum_class=ADPDLed)
        self._config["payload"]["ch1"] = Array(40, dimension=1, data_types=[Int(4)])
        self._config["payload"]["ch2"] = Array(40, dimension=1, data_types=[Int(4)])
        self._config["payload"]["dc0_led_current"] = Int(2)
        self._config["payload"]["tia_ch1"] = Int(2)
        self._config["payload"]["tia_ch2"] = Int(2)


class AgcStatusPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xC',
                'checksum': '0x0'
            },
            'payload': {
                'command': <ADPDCommand.AGC_STATUS_RES: ['0x7D']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'agc_type': <ADPDLed.LED_GREEN: ['0x01']>,
                'agc_status': True
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["agc_type"] = Enums(1, enum_class=ADPDLed)
        self._config["payload"]["agc_status"] = Binary(1)


class ComModePacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xB',
                'checksum': '0x0'
            },
            'payload': {
                'command': <ADPDCommand.COMMUNICATION_MODE_RES: ['0x69']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'com_mode': 2
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["com_mode"] = Int(1)


class SamplingFrequencyPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xC',
                'checksum': '0x0'
            },
            'payload': {
                'command': <ADPDCommand.SET_SAMPLING_FREQUENCY_RES: ['0x73']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'odr': 100
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["odr"] = Int(2)


class SlotPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xF',
                'checksum': '0x0'
            },
            'payload': {
                'command': <ADPDCommand.GET_SLOT_RES: ['0x49']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'slot_num': <ADPDSlot.SLOT_F: ['0x06']>,
                'slot_enabled': True,
                'slot_format': 3,
                'channel_num': 3
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["slot_num"] = Enums(1, enum_class=ADPDSlot)
        self._config["payload"]["slot_enabled"] = Binary(1)
        self._config["payload"]["slot_format"] = Int(2, value_limit=[0, 4])
        self._config["payload"]["channel_num"] = Int(1, value_limit=[1, 3])


class TestCommandPacket(CommandPacket):

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["device_id"] = Array(2, data_types=[Int(1)], default=[0x00, 0x00])
        self._config["payload"]["data"] = Array(12, data_types=[Int(1, to_hex=True)])


class ADPDConfigPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x15',
                'checksum': '0x0'
            },
            'payload': {
                'command': <ADPDCommand.LOAD_CONFIG_RES: ['0x43']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'device_id': <ADPDDevice.DEVICE_GREEN: ['0x28']>,
                'data': [ 255, 255, 31, 0, 112, 7, 0, 99, 0, 0 ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["device_id"] = Enums(1, enum_class=ADPDDevice)
        self._config["payload"]["data"] = Array(10, data_types=[Int(1, to_hex=True)])


class ADPDPauseResumePacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x15',
                'checksum': '0x0'
            },
            'payload': {
                'command': <ADPDCommand.SET_PAUSE_RES: ['0x67']>,
                'status': <CommonStatus.ERROR: ['0x01']>,
                'device_id': <ADPDDevice.DEVICE_G_R_IR_B: ['0x2C']>,
                'pause': True
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["device_id"] = Enums(1, enum_class=ADPDDevice)
        self._config["payload"]["pause"] = Binary(10)


class ClockCalibrationPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xB',
                'checksum': '0x0'
            },
            'payload': {
                'command': <ADPDCommand.CLOCK_CALIBRATION_RES: ['0x45']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'clock_id': <Clock.CLOCK_1M_AND_32M: ['0x06']>
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["clock_id"] = Enums(1, enum_class=Clock)


class ExternalStreamData(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header':{
                'source': <Application.ADPD: ['0xC1','0x10']>,
                'destination': <Application.APP_USB: ['0xC7','0x05']>,
                'length': '0xC',
                'checksum': '0x0'
            },
           'payload':{
                'command': <ADPDCommand.EXT_ADPD_DATA_STREAM: ['0x80']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'sequence_num': [ 2, 0, 0, 0 ],
                'data': [ 37, 26, 6, 0 ],
                'timestamp': [ 148, 201, 188, 2 ]
           }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["sequence_num"] = Int(4)
        self._config["payload"]["data"] = Int(4)
        self._config["payload"]["timestamp"] = Int(4)


class ExternalStreamODR(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xC',
                'checksum': '0x0'
            },
            'payload': {
                'command': <ADPDCommand.SET_EXT_DATA_STREAM_ODR_RES: ['0x7F']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'odr': 50
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["odr"] = Int(2, value_limit=[25, 500])


class ADPDCreateDeviceConfiguration(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 19,
                'checksum': 0
            },
            'payload': {
                'command': <ADPDCommand.CREATE_DCFG_RES: ['0x71']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'size': 2,
                'data': [
                    [ <ADPDSlot.SLOT_A: ['0x01']>, <ADPDAppID.APP_ECG: ['0x00']> ],
                    [ <ADPDSlot.SLOT_B: ['0x02']>, <ADPDAppID.APP_ADPD_GREEN: ['0x04']> ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = Int(1)
        self._config["payload"]["data"] = Array(-1, dimension=2,
                                                data_types=[
                                                    Enums(2, enum_class=ADPDSlot, decode_with_int=True),
                                                    Enums(2, enum_class=ADPDAppID, decode_with_int=True)
                                                ])


class ADPDDCFGPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        [
            {
                'header': {
                    'source': <Application.ADPD: ['0xC1', '0x10']>,
                    'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                    'length': 240,
                    'checksum': 0
                },
                'payload': {
                    'command': <CommonCommand.GET_DCFG_RES: ['0x26']>,
                    'status': <CommonStatus.OK: ['0x00']>,
                    'size': 57,
                    'packet_count': 2,
                    'data': [
                        [ '0x9', '0x97' ],
                        [ '0xB', '0x2E2' ],
                        ...
                        [ '0x1A5', '0x5' ],
                        [ '0x1A6', '0x0' ]
                    ]
                }
            },
            {
                'header': {
                    'source': <Application.ADPD: ['0xC1', '0x10']>,
                    'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                    'length': 240,
                    'checksum': 0
                },
                'payload': {
                    'command': <CommonCommand.GET_DCFG_RES: ['0x26']>,
                    'status': <CommonStatus.OK: ['0x00']>,
                    'size': 10,
                    'packet_count': 2,
                    'data': [
                        [ '0x1A7', '0x120' ],
                        [ '0x1A8', '0x0' ],
                        ...
                        [ '0x1AF', '0x0' ],
                        [ '0x1B0', '0x4' ]
                    ]
                }
            }
        ]
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = Int(1)
        self._config["payload"]["packet_count"] = Int(1)
        self._config["payload"]["data"] = Array(-1, dimension=2,
                                                data_types=[
                                                    Int(2, to_hex=True),
                                                    Int(2, to_hex=True)
                                                ], reverse_inner_array=True)


class ADPDDCBPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        [
            {
                'header': {
                    'source': <Application.ADPD: ['0xC1', '0x10']>,
                    'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                    'length': 242,
                    'checksum': 0
                },
                'payload': {
                    'command': <DCBCommand.READ_CONFIG_RES: ['0x98']>,
                    'status': <DCBStatus.OK: ['0x97']>,
                    'size': 57,
                    'packet_count': 2,
                    'data': [
                        [ '0x9', '0x97' ],
                        [ '0xB', '0x6E2' ],
                        ...
                        [ '0x1A5', '0x5' ],
                        [ '0x1A6', '0x0' ]
                    ]
                }
            },
            {
                'header': {
                    'source': <Application.ADPD: ['0xC1', '0x10']>,
                    'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                    'length': 242,
                    'checksum': 0
                },
                'payload': {
                    'command': <DCBCommand.READ_CONFIG_RES: ['0x98']>,
                    'status': <DCBStatus.OK: ['0x97']>,
                    'size': 10,
                    'packet_count': 2,
                    'data': [
                        [ '0x1A7', '0x120' ],
                        [ '0x1A8', '0x0' ],
                        ...
                        [ '0x1AF', '0x0' ],
                        [ '0x1B0', '0x4' ]
                    ]
                }
            }
        ]
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = Int(2)
        self._config["payload"]["packet_count"] = Int(2)
        self._config["payload"]["data"] = Array(-1, dimension=2,
                                                data_types=[
                                                    Int(2, to_hex=True),
                                                    Int(2, value_limit=[0x0000, 0x0277], to_hex=True)
                                                ], reverse_inner_array=True)


class ADPDRegisterReadPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 23,
                'checksum': 0
            },
            'payload': {
                'command': <CommonCommand.REGISTER_READ_RES: ['0x22']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'size': 3,
                'data': [
                    [ '0x20', '0x2222' ],
                    [ '0x21', '0x0' ],
                    [ '0x22', '0x3' ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = Int(1)
        self._config["payload"]["data"] = Array(-1, dimension=2,
                                                data_types=[
                                                    Int(2, value_limit=[0x0000, 0x0277], to_hex=True),
                                                    Int(2, to_hex=True)
                                                ])


class ADPDRegisterWritePacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 23,
                'checksum': 0
            },
            'payload': {
                'command': <CommonCommand.REGISTER_READ_RES: ['0x22']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'size': 3,
                'data': [
                    [ '0x20', '0x2222' ],
                    [ '0x21', '0x0' ],
                    [ '0x22', '0x3' ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = Int(1)
        self._config["payload"]["data"] = Array(-1, dimension=2,
                                                data_types=[
                                                    Int(2, value_limit=[0x0000, 0x0277], to_hex=True),
                                                    Int(2, to_hex=True)
                                                ])


class ADPDLibraryConfigPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 14,
                'checksum': 0
            },
            'payload': {
                'command': <CommonCommand.WRITE_LCFG_RES: ['0x19']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'size': 1,
                'data': [
                    [ '0x0', '0x1' ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = Int(1)
        self._config["payload"]["data"] = Array(-1, dimension=2,
                                                data_types=[
                                                    Int(1, value_limit=[0x0, 0x0], to_hex=True),
                                                    Int(2, to_hex=True)
                                                ])


class ADPDUCHRPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 13,
                'checksum': 0
            },
            'payload': {
                'command': <ADPDCommand.UC_HR_ENABLE_RES: ['0x83']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'enabled': True,
                'slot': <ADPDSlot.SLOT_F: ['0x06']>
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["enabled"] = Binary(1)
        self._config["payload"]["slot"] = Enums(2, enum_class=ADPDSlot, decode_with_int=True)



from typing import List

from .. import utils
from .data_type import DataType

import logging

logger = logging.getLogger(__name__)


class Binary(DataType):

    def __init__(self, size: int):
        super().__init__(size, False, False)

    def encode(self):
        return utils.split_int_in_bytes(self.get_value(), length=self._size, reverse=self._reverse)

    def decode(self, array: List):
        value = utils.join_multi_length_packets(array, reverse=self._reverse)
        self.set_value(bool(value))
        # none if value not in [0,1]

    def _default_value(self):
        self.set_value(False)

    def _value_check(self, value: int):
        if not type(value) == bool:
            self._default_value()
            logger.error(f"{value} is not of type boolean. Using default value {self.get_value()}.")
            return False

        return True



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

import logging
from typing import List, Dict

from ..core import utils
from .csv_logging import CSVLogger
from .common_stream import CommonStream
from ..core.enums.dcb_enums import DCBCommand
from ..core.packets.common_packets import VersionPacket
from ..core.packets.command_packet import CommandPacket
from ..core.packets.stream_data_packets import EDADataPacket
from ..core.packets.common_packets import DecimationFactorPacket
from ..core.enums.common_enums import Application, Stream, CommonCommand
from ..core.enums.eda_enums import EDACommand, EDADFTWindow, ScaleResistor
from ..core.packets.eda_packets import DynamicScalingPacket, ResistorTIACalibratePacket, EDADCBPacket, \
    EDALibraryConfigPacket, EDADFTPacket, EDARegisterReadPacket, EDARegisterWritePacket, EDABaselineImpedancePacket

logger = logging.getLogger(__name__)


class EDAApplication(CommonStream):
    """
    EDA Application class.

    .. code-block:: python3
        :emphasize-lines: 4

        from adi_study_watch import SDK

        sdk = SDK("COM4")
        application = sdk.get_eda_application()

    """
    DFT_WINDOW_4 = EDADFTWindow.DFT_WINDOW_4
    DFT_WINDOW_8 = EDADFTWindow.DFT_WINDOW_8
    DFT_WINDOW_16 = EDADFTWindow.DFT_WINDOW_16
    DFT_WINDOW_32 = EDADFTWindow.DFT_WINDOW_32

    SCALE_RESISTOR_110 = ScaleResistor.SCALE_RESISTOR_110
    SCALE_RESISTOR_1K = ScaleResistor.SCALE_RESISTOR_1K
    SCALE_RESISTOR_2K = ScaleResistor.SCALE_RESISTOR_2K
    SCALE_RESISTOR_3K = ScaleResistor.SCALE_RESISTOR_3K
    SCALE_RESISTOR_4K = ScaleResistor.SCALE_RESISTOR_4K
    SCALE_RESISTOR_6K = ScaleResistor.SCALE_RESISTOR_6K
    SCALE_RESISTOR_8K = ScaleResistor.SCALE_RESISTOR_8K
    SCALE_RESISTOR_10K = ScaleResistor.SCALE_RESISTOR_10K
    SCALE_RESISTOR_12K = ScaleResistor.SCALE_RESISTOR_12K
    SCALE_RESISTOR_16K = ScaleResistor.SCALE_RESISTOR_16K
    SCALE_RESISTOR_20K = ScaleResistor.SCALE_RESISTOR_20K
    SCALE_RESISTOR_24K = ScaleResistor.SCALE_RESISTOR_24K
    SCALE_RESISTOR_30K = ScaleResistor.SCALE_RESISTOR_30K
    SCALE_RESISTOR_32K = ScaleResistor.SCALE_RESISTOR_32K
    SCALE_RESISTOR_40K = ScaleResistor.SCALE_RESISTOR_40K
    SCALE_RESISTOR_48K = ScaleResistor.SCALE_RESISTOR_48K
    SCALE_RESISTOR_64K = ScaleResistor.SCALE_RESISTOR_64K
    SCALE_RESISTOR_85K = ScaleResistor.SCALE_RESISTOR_85K
    SCALE_RESISTOR_96K = ScaleResistor.SCALE_RESISTOR_96K
    SCALE_RESISTOR_100K = ScaleResistor.SCALE_RESISTOR_100K
    SCALE_RESISTOR_120K = ScaleResistor.SCALE_RESISTOR_120K
    SCALE_RESISTOR_128K = ScaleResistor.SCALE_RESISTOR_128K
    SCALE_RESISTOR_160K = ScaleResistor.SCALE_RESISTOR_160K
    SCALE_RESISTOR_196K = ScaleResistor.SCALE_RESISTOR_196K
    SCALE_RESISTOR_256K = ScaleResistor.SCALE_RESISTOR_256K
    SCALE_RESISTOR_512K = ScaleResistor.SCALE_RESISTOR_512K

    def __init__(self, packet_manager):
        super().__init__(Application.EDA, Stream.EDA, packet_manager)
        self._dcb_size = 2

    @staticmethod
    def get_supported_scales() -> List[ScaleResistor]:
        """
        List all supported scales for EDA.

        :return: Array of scales enums.
        :rtype: List

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.get_supported_scales()
            print(x)
            # [<ScaleResistor.SCALE_RESISTOR_100K: ['0x0', '0x14']>, ...]
        """
        return [EDAApplication.SCALE_RESISTOR_110, EDAApplication.SCALE_RESISTOR_1K, EDAApplication.SCALE_RESISTOR_2K,
                EDAApplication.SCALE_RESISTOR_3K, EDAApplication.SCALE_RESISTOR_4K, EDAApplication.SCALE_RESISTOR_6K,
                EDAApplication.SCALE_RESISTOR_8K, EDAApplication.SCALE_RESISTOR_10K, EDAApplication.SCALE_RESISTOR_12K,
                EDAApplication.SCALE_RESISTOR_16K, EDAApplication.SCALE_RESISTOR_20K, EDAApplication.SCALE_RESISTOR_24K,
                EDAApplication.SCALE_RESISTOR_30K, EDAApplication.SCALE_RESISTOR_32K, EDAApplication.SCALE_RESISTOR_40K,
                EDAApplication.SCALE_RESISTOR_48K, EDAApplication.SCALE_RESISTOR_64K, EDAApplication.SCALE_RESISTOR_85K,
                EDAApplication.SCALE_RESISTOR_96K, EDAApplication.SCALE_RESISTOR_100K,
                EDAApplication.SCALE_RESISTOR_120K, EDAApplication.SCALE_RESISTOR_128K,
                EDAApplication.SCALE_RESISTOR_160K, EDAApplication.SCALE_RESISTOR_196K,
                EDAApplication.SCALE_RESISTOR_256K, EDAApplication.SCALE_RESISTOR_512K]

    @staticmethod
    def get_supported_dft_windows() -> List[EDADFTWindow]:
        """
        List all supported dft window for EDA.

        :return: Array of dft window enums.
        :rtype: List

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.get_supported_dft_windows()
            print(x)
            # [<EDADFTWindow.DFT_WINDOW_4: ['0x0']>, ... , <EDADFTWindow.DFT_WINDOW_32: ['0x3']>]
        """
        return [EDAApplication.DFT_WINDOW_4, EDAApplication.DFT_WINDOW_8, EDAApplication.DFT_WINDOW_16,
                EDAApplication.DFT_WINDOW_32]

    def calibrate_resistor_tia(self, min_scale: ScaleResistor, max_scale: ScaleResistor,
                               lp_resistor_tia: ScaleResistor) -> Dict:
        """
         Calibrate Resistor Trans Impedance Amplifier.

         :param min_scale: min scale for Resistor Trans Impedance Amplifier, use get_supported_scales()
                          | to list all supported scales.
         :param max_scale: max scale for Resistor Trans Impedance Amplifier, use get_supported_scales()
                          | to list all supported scales.
         :param lp_resistor_tia: lp_resistor_tia, use get_supported_scales() to list all supported scales.
         :type min_scale: ScaleResistor
         :type max_scale: ScaleResistor
         :type lp_resistor_tia: ScaleResistor
         :return: A response packet as dictionary.
         :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.get_supported_scales()
            print(x)
            # [<ScaleResistor.SCALE_RESISTOR_100K: ['0x0', '0x14']>, ... ]
            x = application.calibrate_resistor_tia(application.SCALE_RESISTOR_128K, application.SCALE_RESISTOR_256K,
                                                    application.SCALE_RESISTOR_256K)
            print(x["payload"]["status"])
            # CommonStatus.OK
         """
        request_packet = ResistorTIACalibratePacket(self._destination, EDACommand.RESISTOR_TIA_CAL_REQ)
        request_packet.set_payload("min_scale", min_scale)
        request_packet.set_payload("max_scale", max_scale)
        request_packet.set_payload("lp_resistor_tia", lp_resistor_tia)
        size = utils.join_multi_length_packets(max_scale.value) - utils.join_multi_length_packets(min_scale.value) + 1
        request_packet.set_payload("size", size)
        response_packet = ResistorTIACalibratePacket(self._destination, EDACommand.RESISTOR_TIA_CAL_RES)
        return self._send_packet(request_packet, response_packet)

    def get_decimation_factor(self) -> Dict:
        """
        Returns stream decimation factor.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.get_decimation_factor()
            print(x["payload"]["decimation_factor"])
            # 1

        """
        request_packet = DecimationFactorPacket(self._destination, CommonCommand.GET_STREAM_DEC_FACTOR_REQ)
        request_packet.set_payload("stream_address", self._stream)
        response_packet = DecimationFactorPacket(self._destination, CommonCommand.GET_STREAM_DEC_FACTOR_RES)
        return self._send_packet(request_packet, response_packet)

    def set_decimation_factor(self, decimation_factor: int) -> Dict:
        """
        Sets decimation factor for EDA stream.

        :param decimation_factor: decimation factor for stream
        :type decimation_factor: int
        :return: A response packet as dictionary
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Decimation Lower Limit
             - Decimation Upper Limit
           * - 0x01
             - 0x05

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.set_decimation_factor(2)
            print(x["payload"]["decimation_factor"])
            # 2

        """
        request_packet = DecimationFactorPacket(self._destination, CommonCommand.SET_STREAM_DEC_FACTOR_REQ)
        request_packet.set_payload("stream_address", self._stream)
        request_packet.set_payload("decimation_factor", decimation_factor)
        response_packet = DecimationFactorPacket(self._destination, CommonCommand.SET_STREAM_DEC_FACTOR_RES)
        return self._send_packet(request_packet, response_packet)

    def delete_device_configuration_block(self) -> Dict:
        """
        Deletes EDA Device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            application.delete_device_configuration_block()
        """
        request_packet = CommandPacket(self._destination, DCBCommand.ERASE_CONFIG_REQ)
        response_packet = EDADCBPacket(self._destination, DCBCommand.ERASE_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def disable_dynamic_scaling(self) -> Dict:
        """
        Disables Dynamic scaling.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            application.disable_dynamic_scaling()
        """
        request_packet = DynamicScalingPacket(self._destination, EDACommand.DYNAMIC_SCALE_REQ)
        request_packet.set_payload("enabled", 0)
        response_packet = DynamicScalingPacket(self._destination, EDACommand.DYNAMIC_SCALE_RES)
        return self._send_packet(request_packet, response_packet)

    def enable_dynamic_scaling(self, min_scale: ScaleResistor, max_scale: ScaleResistor,
                               lp_resistor_tia: ScaleResistor) -> Dict:
        """
         Enables Dynamic scaling.

         :param min_scale: min scale for Resistor Trans Impedance Amplifier, use get_supported_scales()
                          | to list all supported scales.
         :param max_scale: max scale for Resistor Trans Impedance Amplifier, use get_supported_scales()
                          | to list all supported scales.
         :param lp_resistor_tia: lp_resistor_tia, use get_supported_scales() to list all supported scales.
         :type min_scale: ScaleResistor
         :type max_scale: ScaleResistor
         :type lp_resistor_tia: ScaleResistor
         :return: A response packet as dictionary.
         :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.get_supported_scales()
            print(x)
            # [<ScaleResistor.SCALE_RESISTOR_100K: ['0x0', '0x14']>, ... ]
            x = application.enable_dynamic_scaling(application.SCALE_RESISTOR_128K, application.SCALE_RESISTOR_256K,
                                                    application.SCALE_RESISTOR_256K)
            print(x["payload"]["status"])
            # CommonStatus.OK

         """
        request_packet = DynamicScalingPacket(self._destination, EDACommand.DYNAMIC_SCALE_REQ)
        request_packet.set_payload("enabled", 1)
        request_packet.set_payload("min_scale", min_scale)
        request_packet.set_payload("max_scale", max_scale)
        request_packet.set_payload("lp_resistor_tia", lp_resistor_tia)
        response_packet = DynamicScalingPacket(self._destination, EDACommand.DYNAMIC_SCALE_RES)
        return self._send_packet(request_packet, response_packet)

    def get_version(self) -> Dict:
        """
        Returns EDA version info.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ecg_application()
            x = application.get_version()
            print(x["payload"]["major_version"])
            # 3
            print(x["payload"]["minor_version"])
            # 4
            print(x["payload"]["patch_version"])
            # 3
            print(x["payload"]["version_string"])
            # EDA_App
            print(x["payload"]["build_version"])
            # TEST EDA_VERSION STRING
        """
        request_packet = CommandPacket(self._destination, CommonCommand.GET_VERSION_REQ)
        response_packet = VersionPacket(self._destination, CommonCommand.GET_VERSION_RES)
        return self._send_packet(request_packet, response_packet)

    def read_device_configuration_block(self) -> Dict:
        """
        Returns entire device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.read_device_configuration_block()
            print(x["payload"]["data"])
            # []

        """
        request_packet = CommandPacket(self._destination, DCBCommand.READ_CONFIG_REQ)
        response_packet = EDADCBPacket(self._destination, DCBCommand.READ_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def read_library_configuration(self, fields: List[int]) -> Dict:
        """
        Reads library configuration from specified field values.

        :param fields: List of field values to read.
        :type fields: List[int]
        :return: A response packet as dictionary
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Fields Lower Limit
             - Fields Upper Limit
           * - 0x00
             - 0x02

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.read_library_configuration([0x00])
            print(x["payload"]["data"])
            # [['0x0', '0x0']]

        """
        data = [[field, 0] for field in fields]
        request_packet = EDALibraryConfigPacket(self._destination, CommonCommand.READ_LCFG_REQ)
        request_packet.set_payload("size", len(data))
        request_packet.set_payload("data", data)
        response_packet = EDALibraryConfigPacket(self._destination, CommonCommand.READ_LCFG_RES)
        return self._send_packet(request_packet, response_packet)

    def set_discrete_fourier_transformation(self, dft_window: EDADFTWindow) -> Dict:
        """
        Set Discrete Fourier Transformation for EDA.

        :param dft_window: DFT window for Discrete Fourier Transformation, use get_supported_dft_windows()
                          | to list all supported DFT window.
        :type dft_window: EDADFTWindow
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.get_supported_dft_windows()
            print(x)
            # [<EDADFTWindow.DFT_WINDOW_4: ['0x0']>, ... ,<EDADFTWindow.DFT_WINDOW_32: ['0x3']>]
            x = application.set_discrete_fourier_transformation(application.DFT_WINDOW_32)
            print(x["payload"]["dft_window"])
            # EDADFTWindow.DFT_WINDOW_32
        """
        request_packet = EDADFTPacket(self._destination, EDACommand.SET_DFT_NUM_REQ)
        request_packet.set_payload("dft_window", dft_window)
        response_packet = EDADFTPacket(self._destination, EDACommand.SET_DFT_NUM_RES)
        return self._send_packet(request_packet, response_packet)

    def write_device_configuration_block(self, addresses_values: List[List[int]]) -> Dict:
        """
        Writes the device configuration block values of specified addresses.
        This function takes a list of addresses and values to write, and returns a response packet as
        dictionary containing addresses and values.

        :param addresses_values: List of addresses and values to write.
        :type addresses_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x00
             - 0x02

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.write_device_configuration_block([[0x0, 2], [0x1, 0x1]])
            print(x["payload"]["size"])
            # 2

        """
        request_packet = EDADCBPacket(self._destination, DCBCommand.WRITE_CONFIG_REQ)
        request_packet.set_payload("size", len(addresses_values))
        request_packet.set_payload("data", addresses_values)
        response_packet = EDADCBPacket(self._destination, DCBCommand.WRITE_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def write_device_configuration_block_from_file(self, filename: str) -> Dict:
        """
        Writes the device configuration block values of specified addresses from file.

        :param filename: dcb filename
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x00
             - 0x02

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            application.write_device_configuration_block_from_file("adxl_dcb.dcfg")

        """
        result = self.device_configuration_file_to_list(filename)
        if result:
            return self.write_device_configuration_block(result)

    def write_dcb_to_lcfg(self) -> Dict:
        """
        Writes Device configuration block data to library configuration.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.write_dcb_to_lcfg()
            print(x["payload"]["status"])
            # CommonStatus.OK
        """
        request_packet = CommandPacket(self._destination, CommonCommand.SET_LCFG_REQ)
        response_packet = CommandPacket(self._destination, CommonCommand.SET_LCFG_RES)
        return self._send_packet(request_packet, response_packet)

    def write_library_configuration(self, fields_values: List[List[int]]) -> Dict:
        """
        Writes library configuration from List of fields and values.

        :param fields_values: List of fields and values to write.
        :type fields_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Fields Lower Limit
             - Fields Upper Limit
           * - 0x00
             - 0x02

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.write_library_configuration([[0x00, 0x1]])
            print(x["payload"]["data"])
            # [['0x0', '0x1']]

        """
        request_packet = EDALibraryConfigPacket(self._destination, CommonCommand.WRITE_LCFG_REQ)
        request_packet.set_payload("size", len(fields_values))
        request_packet.set_payload("data", fields_values)
        response_packet = EDALibraryConfigPacket(self._destination, CommonCommand.WRITE_LCFG_RES)
        return self._send_packet(request_packet, response_packet)

    def set_baseline_impedance(self, real_dft16, imaginary_dft16, real_dft8, imaginary_dft8, resistor_baseline) -> Dict:
        """
        Set Baseline Impedance for EDA.

        :param real_dft16: Real DFT16
        :type real_dft16: float
        :param imaginary_dft16: Imaginary DFT16
        :type imaginary_dft16: float
        :param real_dft8: Real DFT8
        :type real_dft8: float
        :param imaginary_dft8: Imaginary DFT8
        :type imaginary_dft8: float
        :param resistor_baseline: Resistor Baseline
        :type resistor_baseline: int
        :return: A response packet as dictionary
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            packet = application.set_baseline_impedance(25000.5, 25000.5, 25000.5, 25000.5, 19900)
            print(packet)
        """
        request_packet = EDABaselineImpedancePacket(self._destination, EDACommand.BASELINE_IMP_SET_REQ)
        request_packet.set_payload("real_dft16", real_dft16)
        request_packet.set_payload("imaginary_dft16", imaginary_dft16)
        request_packet.set_payload("real_dft8", real_dft8)
        request_packet.set_payload("imaginary_dft8", imaginary_dft8)
        request_packet.set_payload("resistor_baseline", resistor_baseline)
        response_packet = EDABaselineImpedancePacket(self._destination, EDACommand.BASELINE_IMP_SET_RES)
        return self._send_packet(request_packet, response_packet)

    def reset_baseline_impedance(self) -> Dict:
        """
        Reset baseline impedance.

        :return: A response packet as dictionary
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.reset_baseline_impedance()
            print(x)
        """
        request_packet = EDABaselineImpedancePacket(self._destination, EDACommand.BASELINE_IMP_RESET_REQ)
        request_packet.set_payload("real_dft16", 0)
        request_packet.set_payload("imaginary_dft16", 0)
        request_packet.set_payload("real_dft8", 0)
        request_packet.set_payload("imaginary_dft8", 0)
        request_packet.set_payload("resistor_baseline", 0)
        response_packet = EDABaselineImpedancePacket(self._destination, EDACommand.BASELINE_IMP_RESET_RES)
        return self._send_packet(request_packet, response_packet)

    def read_register(self, addresses: List[int]) -> Dict:
        """
        Reads the register values of specified addresses. This function takes a list of addresses to read,
        and returns a response packet as dictionary containing addresses and values.

        :param addresses: List of register addresses to read.
        :type addresses: List[int]
        :return: A response packet as dictionary
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x00
             - 0x2E

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.read_register([0x15, 0x20, 0x2E])
            print(x["payload"]["data"])
            # [['0x15', '0x0'], ['0x20', '0x0'], ['0x2E', '0x0']]
        """
        data = [[address, 0] for address in addresses]
        request_packet = EDARegisterReadPacket(self._destination, CommonCommand.REGISTER_READ_32_REQ)
        request_packet.set_payload("size", len(data))
        request_packet.set_payload("data", data)
        response_packet = EDARegisterReadPacket(self._destination, CommonCommand.REGISTER_READ_32_RES)
        return self._send_packet(request_packet, response_packet)

    def write_register(self, addresses_values: List[List[int]]) -> Dict:
        """
        Writes the register values of specified addresses. This function takes a list of addresses and values to write,
        and returns a response packet as dictionary containing addresses and values.

        :param addresses_values: List of register addresses and values to write.
        :type addresses_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x20
             - 0x2E

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.write_register([[0x20, 0x1], [0x21, 0x2], [0x2E, 0x3]])
            print(x["payload"]["data"])
            # [['0x20', '0x1'], ['0x21', '0x2'], ['0x2E', '0x3']]

        """
        request_packet = EDARegisterWritePacket(self._destination, CommonCommand.REGISTER_WRITE_32_REQ)
        request_packet.set_payload("size", len(addresses_values))
        request_packet.set_payload("data", addresses_values)
        response_packet = EDARegisterWritePacket(self._destination, CommonCommand.REGISTER_WRITE_32_RES)
        return self._send_packet(request_packet, response_packet)

    def _callback_data(self, packet, packet_id, callback_function=None, args=None):
        """
        Process and returns the data back to user's callback function.
        """
        self._callback_data_helper(packet, EDADataPacket())

    def enable_csv_logging(self, filename, header=None) -> None:
        """
        Start logging stream data into CSV.

        :param filename: Name of the CSV file.
        :param header: Header list of the CSV file.
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.enable_csv_logging("eda.csv")
        """
        if header is None:
            header = ["Timestamp", "IMP Real(Ohms)", "IMP Img(Ohms)", "IMP Module(Ohms)", "IMP Phase(Rad)",
                      "ADM Real(Ohms)", "ADM Img(Ohms)", "ADM Module(Ohms)", "ADM Phase(Rad)", "Seq No."]
        self._csv_logger[Stream.EDA] = CSVLogger(filename, header)

    def disable_csv_logging(self) -> None:
        """
        Stops logging stream data into CSV.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_eda_application()
            x = application.disable_csv_logging()
        """
        if self._csv_logger.get(Stream.EDA):
            self._csv_logger[Stream.EDA].stop_logging()
        self._csv_logger[Stream.EDA] = None



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************
import os
import logging
from typing import List
from datetime import datetime

import serial
import serial.tools.list_ports
from tqdm import tqdm

from .core import utils
from .core.ble_manager import BLEManager
from .core.enums.board_enums import Board
from .core.enums.fs_enums import FSCommand
from .core.enums.pm_enums import PMCommand
from .core.enums.sqi_enum import SQICommand
from .core.enums.ppg_enums import PPGCommand
from .core.enums.ecg_enums import ECGCommand
from .core.packet_manager import PacketManager
from .application.pm_application import PMApplication
from .core.packets.adpd_packets import ADPDDCFGPacket
from .application.fs_application import FSApplication
from .core.packets.adxl_packets import ADXLDCFGPacket
from .core.packets.common_packets import VersionPacket
from .application.eda_application import EDAApplication
from .application.sqi_application import SQIApplication
from .application.ecg_application import ECGApplication
from .application.bia_application import BIAApplication
from .application.ppg_application import PPGApplication
from .core.enums.pedometer_enums import PedometerCommand
from .application.adxl_application import ADXLApplication
from .application.adpd_application import ADPDApplication
from .application.test_application import TestApplication
from .application.user0_application import User0Application
from .core.packets.ecg_packets import ECGLibraryConfigPacket
from .core.packets.ppg_packets import LibraryConfigDataPacket
from .application.ad7156_application import AD7156Application
from .core.enums.common_enums import Application, CommonCommand
from .core.packets.fs_packets import KeyValuePairResponsePacket
from .application.vsm_mb_sb.pm_application import VSMPMApplication
from .application.low_touch_application import LowTouchApplication
from .application.pedometer_application import PedometerApplication
from .core.packets.pm_packets import DateTimePacket, SystemInfoPacket
from .application.temperature_application import TemperatureApplication

logger = logging.getLogger(__name__)


class SDK:
    """
    SDK class
    """

    STUDY_WATCH = Board.STUDY_WATCH
    VSM_MB_SB = Board.VSM_MB_SB
    __version__ = "4.1.0"

    def __init__(self, serial_port_address: str, mac_address: str = None, baud_rate: int = 921600,
                 board=Board.STUDY_WATCH, logging_filename: str = None, debug: bool = False,
                 ble_vendor_id: int = 0x0456, ble_timeout: int = 5, **kwargs):
        """
        Creates a SDK object

        :param serial_port_address: serial port of the device connected
        :param mac_address: MAC address of the device.
        :param baud_rate: baud rate
        :param board: board to connect (STUDY_WATCH, VSM_MB_SB).
        :param logging_filename: log file name.
        :param debug: control for debug mode.
        """
        self._board = None
        self._mac_address = None
        self._ble_manager = None
        self._serial_object = None
        self._packet_manager = None
        self._alarm_callback = {}
        self.reconnect(serial_port_address, mac_address, baud_rate, board, logging_filename, debug, ble_vendor_id,
                       ble_timeout, **kwargs)

    def reconnect(self, serial_port_address: str, mac_address: str = None, baud_rate: int = 921600,
                  board=Board.STUDY_WATCH, logging_filename: str = None, debug: bool = False,
                  ble_vendor_id: int = 0x0456, ble_timeout: int = 5, **kwargs):
        """
        reconnect method allows you to reconnect to SDK; you must call disconnect before using connect.
        """

        log_format = '%(asctime)s - %(threadName)s - %(name)s - %(levelname)s - %(message)s'
        date_format = '%Y-%m-%d %H:%M:%S'
        if logging_filename:
            logging.basicConfig(filename=logging_filename, filemode='a', format=log_format, datefmt=date_format,
                                level=logging.DEBUG)
        else:
            if debug:
                logging.basicConfig(format=log_format, datefmt=date_format, level=logging.DEBUG)
            else:
                logging.basicConfig(format=log_format, datefmt=date_format)
        logger.debug("----- Study Watch SDK Started -----")
        self._mac_address = mac_address
        self._board = board
        self._serial_object = serial.Serial(serial_port_address, baud_rate, **kwargs)
        self._packet_manager = PacketManager(self._serial_object)
        if self._mac_address:
            self._ble_manager = BLEManager(ble_vendor_id, ble_timeout)
            self._ble_manager.disconnect()
            self._ble_manager.connect(self._mac_address)
            self._packet_manager.set_ble_source()
        else:
            self._packet_manager.set_usb_source()

        self._packet_manager.start_receive_and_process_threads()
        self.get_pm_application().set_datetime(datetime.now())

    def get_adpd_application(self):
        """
        Creates an adpd application object

        :returns: an Adpd Application
        :rtype: ADPDApplication
        """
        return ADPDApplication(self._packet_manager)

    def get_adxl_application(self):
        """
        Creates an adxl application object

        :returns: an Adxl Application
        :rtype: ADXLApplication
        """
        return ADXLApplication(self._packet_manager)

    def get_ecg_application(self):
        """
        Creates an ecg application object

        :returns: an ecg Application
        :rtype: ECGApplication
        """
        return ECGApplication(self._packet_manager)

    def get_eda_application(self):
        """
        Creates an eda application object

        :returns: an eda Application
        :rtype: EDAApplication
        """
        return EDAApplication(self._packet_manager)

    def get_fs_application(self):
        """
        Creates an fs application object

        :returns: an fs Application
        :rtype: FSApplication
        """
        return FSApplication(self._packet_manager)

    def get_pedometer_application(self):
        """
        Creates an pedometer application object

        :returns: an pedometer Application
        :rtype: PedometerApplication
        """
        return PedometerApplication(self._packet_manager)

    def get_pm_application(self):
        """
        Creates an pm application object

        :returns: an pm Application
        :rtype: PMApplication
        """
        if self._board == Board.VSM_MB_SB:
            return VSMPMApplication(self._packet_manager)
        else:
            return PMApplication(self._packet_manager)

    def get_ppg_application(self):
        """
        Creates an ppg application object

        :returns: an Ppg Application
        :rtype: PPGApplication
        """
        return PPGApplication(self._packet_manager)

    def get_temperature_application(self):
        """
        Creates a temperature application object

        :returns: a Temperature Application
        :rtype: TemperatureApplication
        """
        return TemperatureApplication(self._packet_manager)

    def get_sqi_application(self):
        """
        Creates a sqi application object

        :returns: a SQI Application
        :rtype: SQIApplication
        """
        return SQIApplication(self._packet_manager)

    def get_bia_application(self):
        """
        Creates a bia application object

        :returns: a BIA Application
        :rtype: BIAApplication
        """
        return BIAApplication(self._packet_manager)

    def get_ad7156_application(self):
        """
        Creates a ad7156 application object

        :returns: a AD7156 Application
        :rtype: AD7156Application
        """
        return AD7156Application(self._packet_manager)

    def get_low_touch_application(self):
        """
        Creates a low touch application object

        :returns: a LowTouch Application
        :rtype: LowTouchApplication
        """
        return LowTouchApplication(self._packet_manager)

    def get_test_application(self, key_test_callback=None, cap_sense_callback=None):
        """
        Creates a test application object, used for internal firmware testing.

        :returns: a Test Application
        :rtype: TestApplication
        """
        return TestApplication(key_test_callback, cap_sense_callback, self._packet_manager)

    def get_user0_application(self):
        """
        Creates a User0 application object

        :returns: a User Application
        :rtype: User0Application
        """
        return User0Application(self._packet_manager)

    def unsubscribe_all_streams(self):
        """
        Unsubscribe from all application streams
        """
        result = [self.get_adxl_application().unsubscribe_stream(), self.get_sqi_application().unsubscribe_stream(),
                  self.get_ppg_application().unsubscribe_stream(), self.get_bia_application().unsubscribe_stream(),
                  self.get_ecg_application().unsubscribe_stream(), self.get_eda_application().unsubscribe_stream(),
                  self.get_temperature_application().unsubscribe_stream(),
                  self.get_pedometer_application().unsubscribe_stream()]
        adpd_app = self.get_adpd_application()
        fs_app = self.get_fs_application()
        for stream in adpd_app.get_supported_streams():
            result.append(adpd_app.unsubscribe_stream(stream))
        for stream in fs_app.get_supported_streams():
            result.append(fs_app.unsubscribe_stream(stream))
        return result

    @staticmethod
    def _get_packet_id(response_command, destination):
        return utils.join_multi_length_packets(destination.value + response_command.value)

    @staticmethod
    def get_supported_boards():
        return [SDK.STUDY_WATCH, SDK.VSM_MB_SB]

    @staticmethod
    def get_available_ports() -> List:
        """
        returns the list of tuple (port, description, hardware_id) of available ports.
        """
        ports = serial.tools.list_ports.comports()
        result = []
        for port, desc, hardware_id in sorted(ports):
            result.append((port, desc, hardware_id))
        return result

    @staticmethod
    def join_csv(*args, output_filename="combined.csv"):
        """
        Joins multiple data stream csv file into single csv file.
        """
        file_size = os.path.getsize(args[0])
        total_size = 0
        progress_bar = tqdm(total=file_size)
        progress_bar.set_description("ADPD csv join")
        space = {}
        header = []
        all_csv = {}
        first_iter = {}
        result_file = open(output_filename, 'w')

        for file in args:
            csv_file = open(file, 'r')
            header = [csv_file.readline(), csv_file.readline()]
            all_csv[file] = csv_file
            first_iter[file] = True
            space[file] = 0

        # write header
        result_file.write(header[0])
        result_file.write(header[1])

        while True:
            result_line = ""
            empty_file_count = 0
            for file in args:
                line = all_csv[file].readline().strip()
                # checking for empty file condition
                if len(line) == 0:
                    empty_file_count += 1
                    result_line += "," * space[file]
                    continue
                # saving number of elements in a row
                if first_iter[file]:
                    space[file] = len(line.split(","))
                    first_iter[file] = False
                # concat row
                result_line += line + ","
                # progress bar
                if file == args[0]:
                    line_size = len(line.encode('utf-8'))
                    total_size += line_size
                    progress_bar.update(line_size)
            # break condition
            if empty_file_count == len(args):
                break
            result_file.write(result_line + "\n")

        # cleanup
        result_file.close()
        progress_bar.update(file_size - total_size)
        progress_bar.close()
        for file in args:
            all_csv[file].close()

    @staticmethod
    def _csv_write_config(file, config, header):
        file.write(f"Address, Value, {header}\n")
        if not type(config) == list:
            config = [config]
        for data in config:
            for value in data["payload"]["data"]:
                if header == "#PPG_LCFG":
                    file.write(f"{value}\n")
                else:
                    file.write(f"{value[0]}, {value[1]}\n")
        file.write(f"\n")

    @staticmethod
    def _csv_write_version(file, config, header):
        file.write(f"Module, {header}\n")
        file.write(f"Major version, {config['payload']['major_version']}\n")
        file.write(f"Minor version, {config['payload']['minor_version']}\n")
        file.write(f"Patch version, {config['payload']['patch_version']}\n")
        file.write(f"Version string, {config['payload']['version_string']}\n")
        file.write(f"Build version, {config['payload']['build_version']}\n")
        file.write(f"\n")

    # noinspection PyProtectedMember,PyTypeChecker,PyUnresolvedReferences
    @staticmethod
    def convert_log_to_csv(filename):
        """
        Converts M2M2 log file into csv.
        """
        if not os.path.exists(filename):
            raise Exception("File not Found.")

        info_result = {
            "key_value_pair": None, "datetime": None, "system_info": None, "version": None, "ppg_algo_version": None,
            "ped_algo_version": None, "ecg_algo_version": None, "sqi_algo_version": None, "adpd_dcfg": None,
            "adxl_dcfg": None, "ppg_lcfg": None, "ecg_lcfg": None,
        }
        folder_name = filename.split(".")[0]
        try:
            os.mkdir(folder_name)
        except Exception as e:
            logger.debug(e)

        # creating all applications
        packet_manager = PacketManager(None, filename=filename)
        packet_manager.set_usb_source()
        adpd_app = ADPDApplication(packet_manager)
        adxl_app = ADXLApplication(packet_manager)
        ecg_app = ECGApplication(packet_manager)
        eda_app = EDAApplication(packet_manager)
        ped_app = PedometerApplication(packet_manager)
        ppg_app = PPGApplication(packet_manager)
        temp_app = TemperatureApplication(packet_manager)
        sqi_app = SQIApplication(packet_manager)
        bcm_app = BIAApplication(packet_manager)
        pm_app = PMApplication(packet_manager)

        # enabling csv logging
        adxl_app.enable_csv_logging(f"{folder_name}/adxl.csv")
        for i, stream in enumerate(adpd_app.get_supported_streams()):
            if stream == adpd_app.STREAM_STATIC_AGC:
                adpd_app.enable_csv_logging(f"{folder_name}/static_agc.csv", stream=adpd_app.STREAM_STATIC_AGC)
            else:
                adpd_app.enable_csv_logging(f"{folder_name}/adpd{i + 1}.csv", stream=stream)
        ecg_app.enable_csv_logging(f"{folder_name}/ecg.csv")
        eda_app.enable_csv_logging(f"{folder_name}/eda.csv")
        ped_app.enable_csv_logging(f"{folder_name}/ped.csv")
        ppg_app.enable_csv_logging(f"{folder_name}/ppg.csv", stream=ppg_app.STREAM_PPG)
        ppg_app.enable_csv_logging(f"{folder_name}/sync_ppg.csv", stream=ppg_app.STREAM_SYNC_PPG)
        ppg_app.enable_csv_logging(f"{folder_name}/dynamic_agc.csv", stream=ppg_app.STREAM_DYNAMIC_AGC)
        ppg_app.enable_csv_logging(f"{folder_name}/hrv.csv", stream=ppg_app.STREAM_HRV)
        temp_app.enable_csv_logging(f"{folder_name}/temp.csv")
        sqi_app.enable_csv_logging(f"{folder_name}/sqi.csv")
        bcm_app.enable_csv_logging(f"{folder_name}/bia.csv")
        pm_app.enable_csv_logging(f"{folder_name}/adp.csv")

        # subscribing
        apps = [adxl_app, ecg_app, eda_app, ped_app, temp_app, sqi_app, bcm_app]
        for app in apps:
            app._subscribe_stream_data()
        ppg_app._subscribe_stream_data(ppg_app.STREAM_PPG)
        ppg_app._subscribe_stream_data(ppg_app.STREAM_SYNC_PPG)
        ppg_app._subscribe_stream_data(ppg_app.STREAM_DYNAMIC_AGC)
        ppg_app._subscribe_stream_data(ppg_app.STREAM_HRV)
        pm_app._subscribe_stream_data(pm_app.STREAM_BATTERY)
        for stream in adpd_app.get_supported_streams():
            adpd_app._subscribe_stream_data(stream)

        def update_info_dict(packet, response_byte, key):
            packet.decode_packet(response_byte)
            if info_result.get(key) is None:
                info_result[key] = packet.get_dict()
            else:
                if type(info_result[key]) == list:
                    info_result[key] = info_result[key] + [packet.get_dict()]
                else:
                    info_result[key] = [info_result[key], packet.get_dict()]

        # noinspection PyShadowingNames,PyProtectedMember
        def callback(response_byte, packet_id):
            if SDK._get_packet_id(FSCommand.SET_KEY_VALUE_PAIR_RES, Application.FS) == packet_id:
                packet = KeyValuePairResponsePacket()
                update_info_dict(packet, response_byte, "key_value_pair")
            elif SDK._get_packet_id(PMCommand.GET_DATE_TIME_RES, Application.PM) == packet_id:
                packet = DateTimePacket()
                update_info_dict(packet, response_byte, "datetime")
                response_packet = info_result["datetime"]
                dt = datetime(response_packet['payload']['year'], response_packet['payload']['month'],
                              response_packet['payload']['day'], response_packet['payload']['hour'],
                              response_packet['payload']['minute'], response_packet['payload']['second'])
                for _stream in adpd_app.get_supported_streams():
                    adpd_app._update_timestamp(dt, _stream, generate_ts=True)
                for _app in apps:
                    _app._update_timestamp(dt, generate_ts=True)
                ppg_app._update_timestamp(dt, ppg_app.STREAM_PPG, generate_ts=True)
                ppg_app._update_timestamp(dt, ppg_app.STREAM_SYNC_PPG, generate_ts=True)
                ppg_app._update_timestamp(dt, ppg_app.STREAM_DYNAMIC_AGC, generate_ts=True)
                ppg_app._update_timestamp(dt, ppg_app.STREAM_HRV, generate_ts=True)
                pm_app._update_timestamp(dt, pm_app.STREAM_BATTERY, generate_ts=True)

            elif SDK._get_packet_id(PMCommand.SYS_INFO_RES, Application.PM) == packet_id:
                packet = SystemInfoPacket()
                update_info_dict(packet, response_byte, "system_info")
            elif SDK._get_packet_id(CommonCommand.GET_VERSION_RES, Application.PM) == packet_id:
                packet = VersionPacket()
                update_info_dict(packet, response_byte, "version")
            elif SDK._get_packet_id(PPGCommand.GET_ALGO_VENDOR_VERSION_RES, Application.PPG) == packet_id:
                packet = VersionPacket()
                update_info_dict(packet, response_byte, "ppg_algo_version")
            elif SDK._get_packet_id(PedometerCommand.GET_ALGO_VENDOR_VERSION_RES, Application.PEDOMETER) == packet_id:
                packet = VersionPacket()
                update_info_dict(packet, response_byte, "ped_algo_version")
            elif SDK._get_packet_id(ECGCommand.GET_ALGO_VENDOR_VERSION_RES, Application.ECG) == packet_id:
                packet = VersionPacket()
                update_info_dict(packet, response_byte, "ecg_algo_version")
            elif SDK._get_packet_id(SQICommand.GET_ALGO_VENDOR_VERSION_RES, Application.SQI) == packet_id:
                packet = VersionPacket()
                update_info_dict(packet, response_byte, "sqi_algo_version")
            elif SDK._get_packet_id(CommonCommand.GET_DCFG_RES, Application.ADXL) == packet_id:
                packet = ADXLDCFGPacket()
                update_info_dict(packet, response_byte, "adxl_dcfg")
            elif SDK._get_packet_id(CommonCommand.GET_DCFG_RES, Application.ADPD) == packet_id:
                packet = ADPDDCFGPacket()
                update_info_dict(packet, response_byte, "adpd_dcfg")
            elif SDK._get_packet_id(CommonCommand.GET_LCFG_RES, Application.PPG) == packet_id:
                packet = LibraryConfigDataPacket()
                update_info_dict(packet, response_byte, "ppg_lcfg")
            elif SDK._get_packet_id(CommonCommand.READ_LCFG_RES, Application.ECG) == packet_id:
                packet = ECGLibraryConfigPacket()
                update_info_dict(packet, response_byte, "ecg_lcfg")

        initial_packets = [
            [FSCommand.SET_KEY_VALUE_PAIR_RES, Application.FS],
            [PMCommand.GET_DATE_TIME_RES, Application.PM],
            [PMCommand.SYS_INFO_RES, Application.PM],
            [CommonCommand.GET_VERSION_RES, Application.PM],
            [PPGCommand.GET_ALGO_VENDOR_VERSION_RES, Application.PPG],
            [PedometerCommand.GET_ALGO_VENDOR_VERSION_RES, Application.PEDOMETER],
            [ECGCommand.GET_ALGO_VENDOR_VERSION_RES, Application.ECG],
            [SQICommand.GET_ALGO_VENDOR_VERSION_RES, Application.SQI],
            [CommonCommand.GET_DCFG_RES, Application.ADXL],
            [CommonCommand.GET_DCFG_RES, Application.ADPD],
            [CommonCommand.GET_LCFG_RES, Application.PPG],
            [CommonCommand.READ_LCFG_RES, Application.ECG]
        ]
        for initial_packet in initial_packets:
            packet_id = SDK._get_packet_id(initial_packet[0], initial_packet[1])
            packet_manager.subscribe(packet_id, callback)

        # process file
        packet_manager.process_file()

        # disabling csv logging.
        for stream in adpd_app.get_supported_streams():
            adpd_app.disable_csv_logging(stream=stream)
        for app in apps:
            app.disable_csv_logging()
        ppg_app.disable_csv_logging(stream=ppg_app.STREAM_PPG)
        ppg_app.disable_csv_logging(stream=ppg_app.STREAM_SYNC_PPG)
        ppg_app.disable_csv_logging(stream=ppg_app.STREAM_DYNAMIC_AGC)
        ppg_app.disable_csv_logging(stream=ppg_app.STREAM_HRV)
        pm_app.disable_csv_logging()

        # unsubscribing
        apps = [adxl_app, ecg_app, eda_app, ped_app, temp_app, sqi_app, bcm_app]
        for app in apps:
            app._unsubscribe_stream_data()
        ppg_app._unsubscribe_stream_data(ppg_app.STREAM_PPG)
        ppg_app._unsubscribe_stream_data(ppg_app.STREAM_SYNC_PPG)
        ppg_app._unsubscribe_stream_data(ppg_app.STREAM_DYNAMIC_AGC)
        ppg_app._unsubscribe_stream_data(ppg_app.STREAM_HRV)
        pm_app._unsubscribe_stream_data(pm_app.STREAM_BATTERY)
        for stream in adpd_app.get_supported_streams():
            adpd_app._unsubscribe_stream_data(stream)

        with open(f"{folder_name}/Summary.csv", 'w') as f:
            if info_result['key_value_pair']:
                f.write(f"Participant Information, {info_result['key_value_pair']['payload']['value_id']}" + "\n")
                f.write(f"Packet loss, 0" + "\n")

        with open(f"{folder_name}/Configuration.csv", 'w') as f:
            if info_result["adpd_dcfg"]:
                SDK._csv_write_config(f, info_result["adpd_dcfg"], "#ADPD_DCFG")
            if info_result["adxl_dcfg"]:
                SDK._csv_write_config(f, info_result["adxl_dcfg"], "#ADXL_DCFG")
            if info_result["ppg_lcfg"]:
                SDK._csv_write_config(f, info_result["ppg_lcfg"], "#PPG_LCFG")
            if info_result["ecg_lcfg"]:
                SDK._csv_write_config(f, info_result["ecg_lcfg"], "#ECG_LCFG")

        with open(f"{folder_name}/FirmwareVersion.csv", 'w') as f:
            if info_result["version"]:
                SDK._csv_write_version(f, info_result["version"], "PM Firmware Version")
            if info_result["ppg_algo_version"]:
                SDK._csv_write_version(f, info_result["ppg_algo_version"], "PPG Firmware Version")
            if info_result["ped_algo_version"]:
                SDK._csv_write_version(f, info_result["ped_algo_version"], "PED Firmware Version")
            if info_result["ecg_algo_version"]:
                SDK._csv_write_version(f, info_result["ecg_algo_version"], "ECG Firmware Version")
            if info_result["sqi_algo_version"]:
                SDK._csv_write_version(f, info_result["sqi_algo_version"], "SQI Firmware Version")

        with open(f"{folder_name}/PSBoardInfo.csv", 'w') as f:
            if info_result["system_info"]:
                config = info_result["system_info"]
                f.write(f"Module, PS Board Info\n")
                f.write(f"Version, {config['payload']['version']}\n")
                f.write(f"MAC address, {config['payload']['mac_address']}\n")
                f.write(f"Device ID, {config['payload']['device_id']}\n")
                f.write(f"Model number, {config['payload']['model_number']}\n")
                f.write(f"Hardware ID, {config['payload']['hw_id']}\n")
                f.write(f"BOM ID, {config['payload']['bom_id']}\n")
                f.write(f"Batch ID, {config['payload']['batch_id']}\n")
                f.write(f"Board Type, {config['payload']['board_type']}\n")

        # combining adpd csv
        adpd_csv_files = []
        for i in range(1, 13):
            adpd_file = f"{folder_name}/adpd{i}.csv"
            if os.path.exists(adpd_file):
                adpd_csv_files.append(adpd_file)
        if len(adpd_csv_files):
            SDK.join_csv(*adpd_csv_files, output_filename=f"{folder_name}/adpd_streams.csv")

        # cleanup
        del adpd_app
        for _ in apps:
            del _
        del ppg_app
        del pm_app
        del packet_manager

    def disconnect(self):
        """disconnect SDK"""
        logger.debug("----- Study Watch SDK Stopped -----")
        if self._mac_address:
            self._ble_manager.disconnect()
        self._packet_manager.close()



import logging
import time
from queue import Queue

import pywinusb.hid as hid

from . import utils

logger = logging.getLogger(__name__)


class BLEManager:
    RID_CMD = 0x01
    RID_RSP = 0x02
    RID_NOT = 0x03

    CMD_SCAN_STOP = 0x01
    CMD_SCAN_START = 0x00
    CMD_CONNECT = 0x02
    CMD_DISCONNECT = 0x03

    MAX_LENGTH = 63

    def __init__(self, vendor_id, timeout):
        self.queue = Queue()
        self.timeout = timeout
        self.device = hid.HidDeviceFilter(vendor_id=vendor_id).get_devices()[0]
        self.device.set_raw_data_handler(self.callback)

    def callback(self, data):
        logger.debug(f"BLE RX :: {':'.join(utils.convert_int_array_to_hex(data))}")
        self.queue.put(data)

    def _send(self, packet):
        packet = packet + [0 for _ in range(self.MAX_LENGTH - len(packet))]
        logger.debug(f"BLE TX :: {':'.join(utils.convert_int_array_to_hex(packet))}")
        self.report.send(packet)

    def _open(self):
        try:
            self.device.open()
        except Exception as e:
            logger.debug(e)
        self.report = self.device.find_output_reports()[0]

    def _scan_start(self, mac_address):
        default_name = "Nordic_USBD_BLE_UART"
        msg = [self.RID_CMD, self.CMD_SCAN_START, 0, 1] + [ord(i) for i in default_name]
        self._send(msg)
        try:
            while True:
                data = self.queue.get(timeout=self.timeout)
                if data[0] == 3 and data[4:10] == mac_address:
                    break
        except Exception as e:
            logger.debug(e)
            raise Exception(f"Failed to find BLE device {'-'.join(['%X' % i for i in mac_address])}.")

    def _scan_stop(self):
        msg = [self.RID_CMD, self.CMD_SCAN_STOP, 0]
        self._send(msg)
        try:
            while True:
                data = self.queue.get(timeout=self.timeout)
                if data[0] == 2:
                    break
        except Exception as e:
            logger.debug(e)
            raise Exception(f"Failed to stop scan.")

    def connect(self, mac_address):
        self._open()
        mac_address = list(map(int, mac_address.split("-"), [16 for _ in mac_address]))
        mac_address = list(reversed(mac_address))
        self._scan_start(mac_address)
        self._scan_stop()
        msg = [self.RID_CMD, self.CMD_CONNECT] + mac_address
        self._send(msg)
        try:
            while True:
                data = self.queue.get(timeout=self.timeout)
                if data[0] == 3:
                    time.sleep(2)
                    break
        except Exception as e:
            logger.debug(e)
            raise Exception(f"Can't connect to BLE {'-'.join(['%X' % i for i in mac_address])}.")
        logger.info(f"BLE connected to {'-'.join(['%X' % i for i in mac_address])}.")

    def disconnect(self):
        self._open()
        msg = [self.RID_CMD, self.CMD_DISCONNECT]
        self._send(msg)
        self.device.close()
        time.sleep(1)
        self.queue.empty()



import logging
from typing import List

from .data_type import DataType

logger = logging.getLogger(__name__)


class SubStructure(DataType):

    def __init__(self, size: int, reverse: bool = False, default: str = ""):
        super().__init__(size, reverse, default)

    def encode(self):
        return [ord(char) for char in self.get_value()]

    def decode(self, array: List):
        result = ""
        for val in array:
            if val == 0x0:
                break
            result += chr(val)
        self.set_value(str(result))

    def default_value(self):
        return self.set_value("")

    def _value_check(self, value: str):
        if not type(value) == str:
            self.default_value()
            logger.error(f"{value} is not of type str. Using default value {self.get_value()}.")
            return False

        return True



from .command_packet import CommandPacket
from ..data_types.array import Array
from ..data_types.integer import Int


class TemperatureDCFGPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        [
            {
                'header': {
                    'source': <Application.TEMPERATURE: ['0xC3', '0x06']>,
                    'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                    'length': 240,
                    'checksum': 0
                },
                'payload': {
                    'command': <CommonCommand.GET_DCFG_RES: ['0x26']>,
                    'status': <CommonStatus.OK: ['0x00']>,
                    'size': 57,
                    'packet_count': 2,
                    'data': [
                        [ 9, 151 ],
                        [ 11, 738 ],
                        ...
                        [ 421, 5 ],
                        [ 422, 0 ]
                    ]
                }
            },
            {
                'header': {
                    'source': <Application.TEMPERATURE: ['0xC3', '0x06']>,
                    'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                    'length': 240,
                    'checksum': 0
                },
                'payload': {
                    'command': <CommonCommand.GET_DCFG_RES: ['0x26']>,
                    'status': <CommonStatus.OK: ['0x00']>,
                    'size': 10,
                    'packet_count': 2,
                    'data': [
                        [ 423, 288 ],
                        [ 424, 0 ],
                        ..
                        [ 431, 0 ],
                        [ 432, 4 ]
                    ]
                }
            }
        ]
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = Int(1)
        self._config["payload"]["packet_count"] = Int(1)
        self._config["payload"]["data"] = Array(-1, dimension=2, data_types=[Int(2), Int(2)], reverse_inner_array=True)


class TemperatureDCBPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        [
            {
                'header': {
                    'source': <Application.ADPD: ['0xC1', '0x10']>,
                    'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                    'length': 242,
                    'checksum': 0
                },
                'payload': {
                    'command': <DCBCommand.READ_CONFIG_RES: ['0x98']>,
                    'status': <DCBStatus.OK: ['0x97']>,
                    'size': 57,
                    'packet_count': 2,
                    'data': [
                        [ '0x9', '0x97' ],
                        [ '0xB', '0x6E2' ],
                        ...
                        [ '0x1A5', '0x5' ],
                        [ '0x1A6', '0x0' ]
                    ]
                }
            },
            {
                'header': {
                    'source': <Application.ADPD: ['0xC1', '0x10']>,
                    'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                    'length': 242,
                    'checksum': 0
                },
                'payload': {
                    'command': <DCBCommand.READ_CONFIG_RES: ['0x98']>,
                    'status': <DCBStatus.OK: ['0x97']>,
                    'size': 10,
                    'packet_count': 2,
                    'data': [
                        [ '0x1A7', '0x120' ],
                        [ '0x1A8', '0x0' ],
                        ...
                        [ '0x1AF', '0x0' ],
                        [ '0x1B0', '0x4' ]
                    ]
                }
            }
        ]
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = Int(2)
        self._config["payload"]["packet_count"] = Int(2)
        self._config["payload"]["data"] = Array(-1, dimension=1, data_types=[Int(4, to_hex=True)])


class TemperatureLibraryConfigPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PPG: ['0xC3', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 16,
                'checksum': 0
            },
            'payload': {
                'command': <CommonCommand.READ_LCFG_RES: ['0x17']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'field': 1,
                'data': [
                    [ '0x0', '0xC0' ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["field"] = Int(1)
        self._config["payload"]["data"] = Array(-1, dimension=1, data_types=[Int(4, to_hex=True)])



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************
from datetime import datetime
from typing import List, Dict

from ..core import utils
from ..core.enums.dcb_enums import DCBCommand
from .common_application import CommonApplication
from ..core.packets.command_packet import CommandPacket
from ..core.enums.common_enums import Application, CommonCommand
from ..core.enums.user0_enums import User0State, User0Command, User0OperationMode
from ..core.packets.user0_packets import User0LibraryConfigPacket, User0DCBPacket, User0StatePacket, \
    User0ExperimentIDPacket, User0HardwareIDPacket, User0PrevStateEventPacket, BypassUser0TimingPacket


class User0Application(CommonApplication):
    """
    User0 Application class.

    .. code-block:: python3
        :emphasize-lines: 4

        from adi_study_watch import SDK

        sdk = SDK("COM4")
        application = sdk.get_user0_application()

    """

    STATE_SLEEP = User0State.SLEEP
    STATE_ADMIT_STANDBY = User0State.ADMIT_STANDBY
    STATE_END_MONITORING = User0State.END_MONITORING
    STATE_CHARGING_BATTERY = User0State.CHARGING_BATTERY
    STATE_START_MONITORING = User0State.START_MONITORING
    STATE_INTERMITTENT_MONITORING = User0State.INTERMITTENT_MONITORING
    STATE_INTERMITTENT_MONITORING_STOP_LOG = User0State.INTERMITTENT_MONITORING_STOP_LOG
    STATE_INTERMITTENT_MONITORING_START_LOG = User0State.INTERMITTENT_MONITORING_START_LOG
    STATE_OUT_OF_BATTERY_STATE_BEFORE_START_MONITORING = User0State.OUT_OF_BATTERY_STATE_BEFORE_START_MONITORING
    STATE_OUT_OF_BATTERY_STATE_DURING_INTERMITTENT_MONITORING = \
        User0State.OUT_OF_BATTERY_STATE_DURING_INTERMITTENT_MONITORING

    def __init__(self, packet_manager):
        super().__init__(Application.USER0_APP, packet_manager)
        self._dcb_size = 19

    @staticmethod
    def get_supported_states():
        """
        List all supported states for User0.

        :return: Array of user0 state ID enums.
        :rtype: List[User0State]
        """
        return [enum_value for enum_value in User0State]

    def clear_prev_state_event(self):
        """
         Clears the user0 config app's previous state, event received and the corresponding timestamp structure
         maintained registered in the Watch Fw.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_user0_application()
            application.clear_prev_state_event()

        """
        request_packet = CommandPacket(self._destination, User0Command.CLEAR_PREV_ST_EVT_REQ)
        response_packet = CommandPacket(self._destination, User0Command.CLEAR_PREV_ST_EVT_RES)
        return self._send_packet(request_packet, response_packet)

    def get_prev_state_event(self):
        """
        This is a command, to get the user0 config app's previous state, event received and the corresponding timestamp
        registered in the Watch Fw.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_user0_application()
            application.get_prev_state_event()

        """
        request_packet = CommandPacket(self._destination, User0Command.GET_PREV_ST_EVT_REQ)
        response_packet = User0PrevStateEventPacket(self._destination, User0Command.GET_PREV_ST_EVT_RES)
        packet = self._send_packet(request_packet, response_packet)
        date_time = datetime.now()
        ts = (32000.0 * ((date_time.hour * 3600) + (date_time.minute * 60) + date_time.second))
        last_timestamp = [date_time.timestamp(), ts]
        utils.update_timestamp(packet, last_timestamp)
        return packet

    def set_hardware_id(self, value: int):
        """
        Set Hardware ID.

        :param value: Hardware ID.
        :type value: int
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_user0_application()
            application.set_hardware_id(5)

        """
        request_packet = User0HardwareIDPacket(self._destination, User0Command.ID_OP_REQ)
        request_packet.set_payload("operation", User0OperationMode.WRITE)
        request_packet.set_payload("value", value)
        response_packet = User0HardwareIDPacket(self._destination, User0Command.ID_OP_RES)
        return self._send_packet(request_packet, response_packet)

    def get_hardware_id(self):
        """
        Get Hardware ID.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_user0_application()
            application.get_hardware_id()

        """
        request_packet = User0HardwareIDPacket(self._destination, User0Command.ID_OP_REQ)
        request_packet.set_payload("operation", User0OperationMode.READ)
        response_packet = User0HardwareIDPacket(self._destination, User0Command.ID_OP_RES)
        return self._send_packet(request_packet, response_packet)

    def delete_hardware_id(self):
        """
        Delete Hardware ID.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_user0_application()
            application.delete_hardware_id()

        """
        request_packet = User0HardwareIDPacket(self._destination, User0Command.ID_OP_REQ)
        request_packet.set_payload("operation", User0OperationMode.DELETE)
        response_packet = User0HardwareIDPacket(self._destination, User0Command.ID_OP_RES)
        return self._send_packet(request_packet, response_packet)

    def set_experiment_id(self, value: int):
        """
        Set experiment ID.

        :param value: experiment ID.
        :type value: int
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_user0_application()
            application.set_experiment_id(5)

        """
        request_packet = User0ExperimentIDPacket(self._destination, User0Command.ID_OP_REQ)
        request_packet.set_payload("operation", User0OperationMode.WRITE)
        request_packet.set_payload("value", value)
        response_packet = User0ExperimentIDPacket(self._destination, User0Command.ID_OP_RES)
        return self._send_packet(request_packet, response_packet)

    def get_experiment_id(self):
        """
        Get experiment ID.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_user0_application()
            application.get_experiment_id()

        """
        request_packet = User0ExperimentIDPacket(self._destination, User0Command.ID_OP_REQ)
        request_packet.set_payload("operation", User0OperationMode.READ)
        response_packet = User0ExperimentIDPacket(self._destination, User0Command.ID_OP_RES)
        return self._send_packet(request_packet, response_packet)

    def delete_experiment_id(self):
        """
        Remove experiment ID.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_user0_application()
            application.delete_experiment_id()

        """
        request_packet = User0ExperimentIDPacket(self._destination, User0Command.ID_OP_REQ)
        request_packet.set_payload("operation", User0OperationMode.DELETE)
        response_packet = User0ExperimentIDPacket(self._destination, User0Command.ID_OP_RES)
        return self._send_packet(request_packet, response_packet)

    def get_state(self):
        """
        Get User0 state.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_user0_application()
            application.get_state()

        """
        request_packet = CommandPacket(self._destination, User0Command.GET_STATE_REQ)
        response_packet = User0StatePacket(self._destination, User0Command.GET_STATE_RES)
        return self._send_packet(request_packet, response_packet)

    def set_state(self, state: User0State):
        """
        Set User0 state.

        :param state: User0 state.
        :type state: User0State
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_user0_application()
            application.set_state(application.STATE_SLEEP)

        """
        request_packet = User0StatePacket(self._destination, User0Command.SET_STATE_REQ)
        request_packet.set_payload("state", state)
        response_packet = User0StatePacket(self._destination, User0Command.SET_STATE_RES)
        return self._send_packet(request_packet, response_packet)

    def read_library_configuration(self, fields: List[int]) -> Dict:
        """
        Reads library configuration from specified field values.

        :param fields: List of field values to read.
        :type fields: List[int]
        :return: A response packet as dictionary
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Fields Lower Limit
             - Fields Upper Limit
           * - 0x00
             - 0x13

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_user0_application()
            x = application.read_library_configuration([0x00])
            print(x["payload"]["data"])
            # [['0x0', '0x0']]
        """
        data = [[field, 0] for field in fields]
        request_packet = User0LibraryConfigPacket(self._destination, CommonCommand.READ_LCFG_REQ)
        request_packet.set_payload("size", len(data))
        request_packet.set_payload("data", data)
        response_packet = User0LibraryConfigPacket(self._destination, CommonCommand.READ_LCFG_RES)
        return self._send_packet(request_packet, response_packet)

    def write_library_configuration(self, fields_values: List[List[int]]) -> Dict:
        """
        Writes library configuration from List of fields and values.

        :param fields_values: List of fields and values to write.
        :type fields_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Fields Lower Limit
             - Fields Upper Limit
           * - 0x00
             - 0x13

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_user0_application()
            x = application.write_library_configuration([[0x01, 0x2]])
            print(x["payload"]["data"])
            # [['0x0', '0x1']]

        """
        request_packet = User0LibraryConfigPacket(self._destination, CommonCommand.WRITE_LCFG_REQ)
        request_packet.set_payload("size", len(fields_values))
        request_packet.set_payload("data", fields_values)
        response_packet = User0LibraryConfigPacket(self._destination, CommonCommand.WRITE_LCFG_RES)
        return self._send_packet(request_packet, response_packet)

    def delete_device_configuration_block(self) -> Dict:
        """
        Delete device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_user0_application()
            application.delete_device_configuration_block()

        """
        request_packet = CommandPacket(self._destination, DCBCommand.ERASE_CONFIG_REQ)
        response_packet = User0DCBPacket(self._destination, DCBCommand.ERASE_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def read_device_configuration_block(self) -> Dict:
        """
        Returns entire device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_user0_application()
            x = application.read_device_configuration_block()
            print(x["payload"]["data"])
            # []

        """
        request_packet = CommandPacket(self._destination, DCBCommand.READ_CONFIG_REQ)
        response_packet = User0DCBPacket(self._destination, DCBCommand.READ_CONFIG_RES)
        response_dict = self._send_packet(request_packet, response_packet)
        response_dict["payload"]["data"] = utils.add_index_to_array(response_dict["payload"]["data"], to_hex=True)
        return response_dict

    def write_device_configuration_block(self, values: List[int]) -> Dict:
        """
        Writes the device configuration block values of specified addresses.
        This function takes a list of addresses and values to write, and returns a response packet as
        dictionary containing addresses and values.

        :param values: List of addresses and values to write.
        :type values: List[int]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x00
             - 0x13

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_user0_application()
            x = application.write_device_configuration_block([0x10, 0x01])
            print(x["payload"]["size"])
            # 2
        """
        request_packet = User0DCBPacket(self._destination, DCBCommand.WRITE_CONFIG_REQ)
        request_packet.set_payload("size", len(values))
        request_packet.set_payload("data", values)
        response_packet = User0DCBPacket(self._destination, DCBCommand.WRITE_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def write_device_configuration_block_from_file(self, filename: str) -> Dict:
        """
        Writes the device configuration block values of specified addresses from file.

        :param filename: dcb filename
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x00
             - 0x13

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_user0_application()
            application.write_device_configuration_block_from_file("user0_dcb.lcfg")
        """
        result = self.device_configuration_file_to_list(filename, address=False)
        if result:
            return self.write_device_configuration_block(result)

    def enable_user0_bypass_timings(self) -> Dict:
        """
        Enable user0 bypass timings.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_user0_application()
            application.enable_user0_bypass_timings()

        """
        request_packet = BypassUser0TimingPacket(self._destination, User0Command.BYPASS_USER0_TIMINGS_REQ)
        request_packet.set_payload("enabled", True)
        response_packet = BypassUser0TimingPacket(self._destination, User0Command.BYPASS_USER0_TIMINGS_RES)
        return self._send_packet(request_packet, response_packet)

    def disable_user0_bypass_timings(self) -> Dict:
        """
        Disable user0 bypass timings.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_user0_application()
            application.disable_user0_bypass_timings()

        """
        request_packet = BypassUser0TimingPacket(self._destination, User0Command.BYPASS_USER0_TIMINGS_REQ)
        request_packet.set_payload("enabled", False)
        response_packet = BypassUser0TimingPacket(self._destination, User0Command.BYPASS_USER0_TIMINGS_RES)
        return self._send_packet(request_packet, response_packet)



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from ..data_types.array import Array
from ..data_types.binary import Binary
from ..data_types.enums import Enums
from ..data_types.integer import Int
from .command_packet import CommandPacket
from ..enums.user0_enums import User0State, User0ID, User0OperationMode, User0Event


class User0LibraryConfigPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.USER0_APP: ['0xC8', '0x0F']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 16,
                'checksum': 0
            },
            'payload': {
                'command': <CommonCommand.READ_LCFG_RES: ['0x17']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'size': 1,
                'data': [
                    [ '0x0', '0x42' ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = Int(1)
        self._config["payload"]["data"] = Array(-1, dimension=2,
                                                data_types=[
                                                    Int(1, value_limit=[0x00, 0x12], to_hex=True),
                                                    Int(4, to_hex=True)
                                                ])


class User0StatePacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.USER0_APP: ['0xC8', '0x0F']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 11,
                'checksum': 0
            },
            'payload': {
                'command': <User0Command.GET_STATE_RES: ['0x45']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'state': <User0State.END_MONITORING: ['0x06']>
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["state"] = Enums(1, enum_class=User0State)


class User0DCBPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.USER0_APP: ['0xC8', '0x0F']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 88,
                'checksum': 0
            },
            'payload': {
                'command': <DCBCommand.READ_CONFIG_RES: ['0x98']>,
                'status': <DCBStatus.OK: ['0x97']>,
                'size': 19,
                'data': [
                    [ '0x00', '0x42' ],
                    [ '0x01', '0x1' ],
                    ...
                    [ '0x11', '0xF' ],
                    [ '0x12', '0x249F0' ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = Int(2)
        self._config["payload"]["data"] = Array(-1, dimension=1, data_types=[Int(4, to_hex=True)], size_limit=0x13)


class User0HardwareIDPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.USER0_APP: ['0xC8', '0x0F']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 14,
                'checksum': 0
            },
            'payload': {
                'command': <User0Command.ID_OP_RES: ['0x47']>,
                'status': <User0Status.OK: ['0x41']>,
                'id': <User0ID.HW_ID: ['0x00']>,
                'operation': <User0OperationMode.READ: ['0x00']>,
                'value': 34
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["id"] = Enums(1, enum_class=User0ID, default=User0ID.HW_ID)
        self._config["payload"]["operation"] = Enums(1, enum_class=User0OperationMode)
        self._config["payload"]["value"] = Int(2, value_limit=[1, 99])


class BypassUser0TimingPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.USER0_APP: ['0xC8', '0x0F']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 14,
                'checksum': 0
            },
            'payload': {
                'command': <User0Command.ID_OP_RES: ['0x47']>,
                'status': <User0Status.OK: ['0x41']>,
                'enabled': false
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["enabled"] = Binary(1)


class User0ExperimentIDPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.USER0_APP: ['0xC8', '0x0F']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 14,
                'checksum': 0
            },
            'payload': {
                'command': <User0Command.ID_OP_RES: ['0x47']>,
                'status': <User0Status.OK: ['0x41']>,
                'id': <User0ID.EXP_ID: ['0x01']>,
                'operation': <User0OperationMode.READ: ['0x00']>,
                'value': 9999
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["id"] = Enums(1, enum_class=User0ID, default=User0ID.EXP_ID)
        self._config["payload"]["operation"] = Enums(1, enum_class=User0OperationMode)
        self._config["payload"]["value"] = Int(2, value_limit=[1, 9999])


class User0PrevStateEventPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.USER0_APP: ['0xC8', '0x0F']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 36,
                'checksum': 0
            },
            'payload': {
                'command': <User0Command.GET_PREV_ST_EVT_RES: ['0x4B']>,
                'status': <User0Status.OK: ['0x41']>,
                'intermittent_count': 1,
                'stream_data': [
                    {
                        'state': <User0State.END_MONITORING: ['0x06']>,
                        'event': <User0Event.WATCH_ON_CRADLE_NAV_BUTTON_RESET: ['0x02']>,
                        'timestamp': 1633154418511.9675
                    },
                    {
                        'state': <User0State.END_MONITORING: ['0x06']>,
                        'event': <User0Event.FINISH_LOG_TRANSFER: ['0x0B']>,
                        'timestamp': 1633154737959.6863
                    },
                    {
                        'state': <User0State.END_MONITORING: ['0x06']>,
                        'event': <User0Event.WATCH_ON_CRADLE_NAV_BUTTON_RESET: ['0x02']>,
                        'timestamp': 1633161548514.8425
                    },
                    {
                        'state': <User0State.END_MONITORING: ['0x06']>,
                        'event': <User0Event.WATCH_ON_CRADLE_NAV_BUTTON_RESET: ['0x02']>,
                        'timestamp': 1633229664514.8113
                    }
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["prev_state_1"] = Enums(1, enum_class=User0State)
        self._config["payload"]["prev_event_1"] = Enums(1, enum_class=User0Event)
        self._config["payload"]["prev_timestamp_1"] = Int(4)
        self._config["payload"]["prev_state_2"] = Enums(1, enum_class=User0State)
        self._config["payload"]["prev_event_2"] = Enums(1, enum_class=User0Event)
        self._config["payload"]["prev_timestamp_2"] = Int(4)
        self._config["payload"]["prev_state_3"] = Enums(1, enum_class=User0State)
        self._config["payload"]["prev_event_3"] = Enums(1, enum_class=User0Event)
        self._config["payload"]["prev_timestamp_3"] = Int(4)
        self._config["payload"]["prev_state_4"] = Enums(1, enum_class=User0State)
        self._config["payload"]["prev_event_4"] = Enums(1, enum_class=User0Event)
        self._config["payload"]["prev_timestamp_4"] = Int(4)
        self._config["payload"]["intermittent_count"] = Int(2)

    def get_dict(self, last_timestamp=None):
        """
        Reorganising of data.
        """
        packet = super().get_dict()
        packet["payload"]["stream_data"] = []
        for i in range(1, 5):
            data = {
                "state": packet["payload"][f"prev_state_{i}"],
                "event": packet["payload"][f"prev_event_{i}"],
                "timestamp": packet["payload"][f"prev_timestamp_{i}"],
            }
            [packet["payload"].pop(key) for key in [f"prev_timestamp_{i}", f"prev_event_{i}", f"prev_state_{i}"]]
            packet["payload"]["stream_data"].append(data)
        return packet
